package resolvers

import (
	"reflect"

	"github.com/graphql-go/graphql"

	account "gitlab.okta-solutions.com/mashroom/backend/account/api"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/generic"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	document "gitlab.okta-solutions.com/mashroom/backend/document/api"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	maintenance "gitlab.okta-solutions.com/mashroom/backend/maintenance/api"
	offer "gitlab.okta-solutions.com/mashroom/backend/offer/api"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
)

func initAccountEnrichment() {
	util.BindOutput(&account.AccountDetails{}, "", util.EnrichmentMap{
		"avatar": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&document.File{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*account.AccountDetails)
				if !ok || acc.GetAvatarValue() == "" {
					return &document.File{}, nil
				}
				resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
					Id: acc.GetAvatarValue(),
				})
				if err != nil {
					return &document.File{}, nil
				}
				return resp.File, err
			},
		},
	})

	util.BindOutput(&account.Account{}, "", util.EnrichmentMap{
		"id": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*account.Account)
				if !ok {
					return "", nil
				}

				return acc.GetIdValue(), nil
			},
		},
		"documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				acc, ok := p.Source.(*account.Account)
				if !ok || acc.GetIdValue() == "" {
					return []*document.Document{}, nil
				}
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*document.Document{}, nil
				}
				switch {
				case acc.GetIdValue() == user.GetAccount().GetIdValue():
				default:
					return []*document.Document{}, nil
				}

				resp, err := connections.DocumentService.List(p.Context, &document.ListRequest{
					User: acc.GetIdValue(),
				})
				if err != nil {
					return []*document.Document{}, nil
				}
				var docs []*document.Document
				for _, d := range resp.Documents {
					if d.Type == document.DocumentType_IDENTITY {
						docs = append(docs, d)
					}
				}
				return docs, err
			},
		},
	})
}

func initPropertyEnrichment() {
	resolveFileList := func(resolver func(desc *property.PropertyDescription) *generic.StringArrayValue) *util.Enrichment {
		return &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.File{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyDescription)
				if !ok {
					return []*document.File{}, nil
				}
				arr := resolver(prop)
				if arr == nil || len(arr.GetValue()) == 0 {
					return []*document.File{}, nil
				}
				var ids []*document.File
				for _, id := range arr.GetValue() {
					resp, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{
						Id: id,
					})
					if err != nil {
						continue
					}
					ids = append(ids, resp.GetFile())
				}
				return ids, nil
			},
		}
	}

	util.BindOutput(&property.PropertyDescription{}, "Property-related descriptions", util.EnrichmentMap{
		"photos": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetPhotos()
		}),
		"videos": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetVideos()
		}),
		"plans": resolveFileList(func(desc *property.PropertyDescription) *generic.StringArrayValue {
			return desc.GetPlans()
		}),
	})

	util.BindOutput(&property.Tenant{}, "Property-related tenant", util.EnrichmentMap{
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Tenant)
				if !ok || prop.GetUser().GetValue() == "" {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: prop.GetUser().GetValue()})
				if err != nil {
					return &account.Account{}, err
				}
				return CensoredAccount(acc.GetAccount()), nil
			},
		},
	})

	util.BindOutput(&property.PropertyPeople{}, "Property-related people", util.EnrichmentMap{
		"landlord": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetLandlord().GetValue() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: prop.GetLandlord().GetValue()})
				if err != nil {
					return nil, err
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), err
			},
		},
		"landlord_invitation": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.InvitationStatus{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return nil, nil
				}

				return prop.GetLandlordInvitation(), nil
			},
		},
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetManager().GetValue() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: prop.GetManager().GetValue()})
				if err != nil {
					return nil, err
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), err
			},
		},
		"manager_invitation": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.InvitationStatus{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return nil, nil
				}

				return prop.GetManagerInvitation(), nil
			},
		},
		"submitter": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok || prop.GetSubmitter().GetValue() == "" {
					return nil, nil
				}
				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: prop.GetSubmitter().GetValue()})
				if err != nil {
					return nil, err
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignProperty(user, prop) {
					return CensoredAccount(acc.GetAccount()), nil
				}

				return acc.GetAccount(), err
			},
		},
		"tenants": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.Tenant{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.PropertyPeople)
				if !ok {
					return []*property.Tenant{}, nil
				}

				return prop.GetTenants().GetValue(), nil
			},
		},
	})

	util.BindOutput(&property.Viewing{}, "Property-related people", util.EnrichmentMap{
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "manager", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.Viewing)
				if !ok {
					return &account.Account{}, nil
				}
				if req.GetManager() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetManager()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutput(&property.Property{}, "Property struct", util.EnrichmentMap{
		"id": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Property)
				if !ok {
					return "", nil
				}

				return prop.GetId().GetValue(), nil
			},
		},
		"favorited": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.Boolean
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return false, nil
				}
				prop, ok := p.Source.(*property.Property)
				if !ok {
					return false, nil
				}

				for _, p := range user.GetAccount().GetFavoritesValue().GetProperties().GetValue() {
					if p == prop.GetId().GetValue() {
						return true, nil
					}
				}
				return false, nil
			},
		},
		"documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Property)
				if !ok || prop.GetId().GetValue() == "" {
					return []*document.Document{}, nil
				}
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return []*document.Document{}, nil
				}

				switch {
				case prop.GetPeople().GetLandlord().GetValue() == user.GetAccount().GetIdValue():
				case prop.GetPeople().GetSubmitter().GetValue() == user.GetAccount().GetIdValue():
				case prop.GetPeople().GetManager().GetValue() == user.GetAccount().GetIdValue():
				default:
					return []*document.Document{}, nil
				}

				resp, err := connections.DocumentService.List(p.Context, &document.ListRequest{
					Property: prop.GetId().GetValue(),
				})
				if err != nil {
					return []*document.Document{}, nil
				}
				return resp.GetDocuments(), err
			},
		},
		"offers_list": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&offer.OfferResponse{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return false, nil
				}

				prop, ok := p.Source.(*property.Property)
				if !ok || prop.GetId().GetValue() == "" {
					return []*offer.OfferResponse{}, nil
				}

				if user.GetAccount().GetIdValue() == "" {
					return []*offer.OfferResponse{}, nil
				}

				var err error

				var resp *offer.ListbyPropertyResponse

				switch user.GetAccount().GetIdValue() {
				case prop.GetPeople().GetLandlord().GetValue(), prop.GetPeople().GetManager().GetValue(), prop.GetPeople().GetSubmitter().GetValue():
					resp, err = connections.OfferService.ListByPropertyLandlord(p.Context, &offer.ListRequest{
						Property: &offer.ListRequest_PropertyValue{PropertyValue: prop.GetId().GetValue()},
						User:     user.GetAccount().GetIdValue(),
					})
				default:
					resp, err = connections.OfferService.ListByPropertyTenant(p.Context, &offer.ListRequest{
						Property: &offer.ListRequest_PropertyValue{PropertyValue: prop.GetId().GetValue()},
						User:     user.GetAccount().GetIdValue(),
					})
				}

				if err != nil {
					return []*offer.OfferResponse{}, nil
				}

				return resp.Offers, nil
			},
		},
		"viewings_list": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.Viewing{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Property)
				if !ok || prop.GetId().GetValue() == "" {
					return []*property.Viewing{}, nil
				}
				resp, err := connections.PropertyService.ListAllView(p.Context, &property.ListViewRequest{
					Property: prop.GetId(),
				})
				if err != nil {
					return []*property.Viewing{}, nil
				}
				return resp.List, err
				// user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				// if !ok || user == nil {
				//	return []*offer.OfferResponse{}, nil
				// }
				// userID := user.GetAccount().GetIdValue()
				// switch userID {
				// case prop.GetPeopleValue().GetLandlordValue(), prop.GetPeopleValue().GetManagerValue(),
				//	prop.GetPeopleValue().GetSubmitterValue():
				//	resp, err := connections.PropertyService.ListAllView(p.Context, &property.ListViewRequest{
				//		Property: &property.ListViewRequest_PropertyValue{PropertyValue: prop.GetIdValue()},
				//	})
				//	if err != nil {
				//		return []*property.Viewing{}, nil
				//	}
				//	return resp.List, err
				// default:
				//	resp, err := connections.PropertyService.ListAllView(p.Context, &property.ListViewRequest{
				//		Property: &property.ListViewRequest_PropertyValue{PropertyValue: prop.GetIdValue()},
				//		Sender:   &property.ListViewRequest_SenderValue{SenderValue: userID},
				//	})
				//	if err != nil {
				//		return []*property.Viewing{}, nil
				//	}
				//	return resp.List, err
				// }
			},
		},
		"status_history": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.PropertyStatusData{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*property.Property)
				if !ok || prop.GetId().GetValue() == "" {
					return []*property.PropertyStatusData{}, nil
				}

				resp, err := connections.PropertyService.GetPropertyStatusHistory(p.Context, &property.PropertyStatusHistoryRequest{
					Property: prop.GetId().GetValue(),
				})
				if err != nil {
					return []*property.PropertyStatusData{}, nil
				}
				return resp.History, err
			},
		},
	})

	util.BindOutput(&property.ViewRequestFeedbackEntry{}, "Property-related details entry", util.EnrichmentMap{
		"initiator": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequestFeedbackEntry)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetInitiator() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetInitiator()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutput(&property.ViewRequestDetails{}, "Property-related details", util.EnrichmentMap{
		"flatmates": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&account.Account{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequestDetails)
				if !ok {
					return []*account.Account{}, nil
				}

				if len(req.GetFlatmates().GetValue()) == 0 {
					return []*account.Account{}, nil
				}

				var list []*account.Account
				for _, f := range req.GetFlatmates().GetValue() {
					acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: f})
					if err != nil {
						return nil, err
					}
					list = append(list, acc.GetAccount())
				}
				return list, nil
			},
		},
	})

	util.BindOutput(offer.Guarantor{}, "offerGuarantor", util.EnrichmentMap{
		"identity_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Guarantor)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document

				if req.GetIdentityDocuments() == nil {
					req.IdentityDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetIdentityDocuments().GetValue() {
					doc, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					if doc.GetSuccess() {
						d := &document.Document{
							Id:   id,
							Type: document.DocumentType_IDENTITY,
							File: doc.GetFile(),
						}
						documents = append(documents, d)
					}
				}
				return documents, nil
			},
		},
		"bill_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Guarantor)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetBillDocuments() == nil {
					req.BillDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetBillDocuments().GetValue() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
	})

	util.BindOutput(offer.Tenant{}, "offerTenant", util.EnrichmentMap{
		"guarantor": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Guarantor{}, "offerGuarantor", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return nil, nil
				}

				return req.GetGuarantor(), nil
			},
		},
		"identity_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetIdentityDocuments() == nil {
					req.IdentityDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetIdentityDocuments().GetValue() {
					doc, err := connections.DocumentService.FileInfo(p.Context, &document.FileInfoRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					if doc.GetSuccess() {
						d := &document.Document{
							Id:   id,
							Type: document.DocumentType_IDENTITY,
							File: doc.GetFile(),
						}
						documents = append(documents, d)
					}
				}
				return documents, nil
			},
		},
		"bill_documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				if req.GetBillDocuments() == nil {
					req.BillDocuments = &generic.StringArrayValue{}
				}
				for _, id := range req.GetBillDocuments().GetValue() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
		"account": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Tenant)
				if !ok {
					return &account.Account{}, nil
				}
				if req.GetUser() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetUser()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
	})

	util.BindOutput(&property.ViewRequest{}, "Property-related people", util.EnrichmentMap{
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil {
					return &property.Property{}, nil
				}

				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &property.Property{}, nil
				}
				if req.GetId() == "" {
					return &property.Property{}, nil
				}
				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{Id: req.GetProperty(), User: user.GetAccount().GetIdValue()})
				if err != nil {
					return nil, err
				}
				return resp.GetProperty(), nil
			},
		},
		"sender": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &account.Account{}, nil
				}

				if req.GetSender() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetSender()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
		"offer": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&offer.Offer{}, "offer", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &offer.Offer{}, nil
				}

				if req.GetSender() == "" {
					return &offer.Offer{}, nil
				}
				resp, err := connections.OfferService.Get(p.Context, &offer.GetRequest{Id: req.GetOffer()})
				if err != nil {
					return &offer.Offer{}, nil
				}
				return resp.GetOffer(), err
			},
		},
		"manager": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "manager", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*property.ViewRequest)
				if !ok {
					return &account.Account{}, nil
				}
				if req.GetManager() == "" {
					return &account.Account{}, nil
				}
				resp, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: req.GetManager()})
				if err != nil {
					return &account.Account{}, nil
				}
				return resp.GetAccount(), err
			},
		},
	})

	delete(util.DecomposedTypes, reflect.TypeOf(&offer.Offer{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.OfferConditions{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.Landlord{}))
	delete(util.DecomposedTypes, reflect.TypeOf(&offer.OfferDetails{}))
}

func initOfferEnrichment() {
	util.BindOutput(offer.Offer{}, "offer", util.EnrichmentMap{
		"viewing": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.ViewRequest{}, "viewing", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if req.GetTenants() == nil {
					req.Tenants = &offer.TenantArrayValue{}
				}
				if !ok || len(req.GetTenants().GetValue()) < 1 {
					return &property.ViewRequest{}, nil
				}
				view, err := connections.PropertyService.GetView(p.Context, &property.GetViewRequest{
					User:     req.GetTenants().GetValue()[0].GetUser(),
					Property: req.GetProperty(),
					Id:       req.GetViewing(),
				})
				if err != nil || view == nil || view.GetRequest() == nil {
					return &property.ViewRequest{}, nil
				}
				return view.GetRequest(), nil
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{Id: req.GetProperty()})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
	})
}

func initMaintenanceEnrichment() {
	util.BindOutput(maintenance.Maintenance{}, "maintenance", util.EnrichmentMap{
		"documents": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&document.Document{}, "document", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return []*document.Document{}, nil
				}

				var documents []*document.Document
				for _, id := range req.GetDocumentsValue().GetValue() {
					doc, err := connections.DocumentService.Get(p.Context, &document.GetRequest{Id: id})
					if err != nil {
						log.WithError(err)
						continue
					}
					documents = append(documents, doc.GetDocument())
				}

				return documents, nil
			},
		},
	})
}

func InitSchema(rootQueries, rootMutations, rootSubscriptions *graphql.Fields) {
	initAccountEnrichment()
	initPropertyEnrichment()
	initOfferEnrichment()
	initMaintenanceEnrichment()

	initAccount(rootQueries, rootMutations)
	initAddress(rootQueries, rootMutations)
	initAuth(rootQueries, rootMutations)
	initCardPayment(rootQueries, rootMutations)
	initContact(rootQueries, rootMutations)
	initContract(rootQueries, rootMutations)
	initDocument(rootQueries, rootMutations)
	initGeo(rootQueries, rootMutations)
	initInvitation(rootQueries, rootMutations)
	initMaintenance(rootQueries, rootMutations)
	initMessaging(rootQueries, rootMutations, rootSubscriptions)
	initOAuth(rootQueries, rootMutations)
	initOffer(rootQueries, rootMutations)
	initPayment(rootQueries, rootMutations)
	initProperty(rootQueries, rootMutations)
	initService(rootQueries, rootMutations)
	initSignature(rootQueries, rootMutations)
	// initVerification(rootQueries, rootMutations)
}
