package admin

import (
	"github.com/graphql-go/graphql"
	account "gitlab.okta-solutions.com/mashroom/backend/account/api"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	q "gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	maintenance "gitlab.okta-solutions.com/mashroom/backend/maintenance/api"
	offer "gitlab.okta-solutions.com/mashroom/backend/offer/api"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
)

const (
	MediaCacheDirEnv     = "MEDIA_CACHE_DIR"
	MediaCacheDirDefault = "/media"
)

var MediaCacheDir = env.ResolveString(MediaCacheDirEnv, MediaCacheDirDefault)

type allParamsPagination struct {
	Page    int32 `json:"page,omitempty"`
	PerPage int32 `json:"perPage,omitempty"`
}

type allParams struct {
	Pagination allParamsPagination    `json:"pagination,omitempty"`
	Sort       []*q.QueryParamsSort   `json:"sort,omitempty"`
	Filter     []*q.QueryParamsFilter `json:"filter,omitempty"`
}

type getParams struct {
	ID string `json:"id"`
}

func InitSchema(rootQueries, rootMutations *graphql.Fields) {
	queryFields := graphql.Fields{}
	mutationFields := graphql.Fields{}

	initUser(&queryFields, &mutationFields)
	initAdmin(&queryFields, &mutationFields)
	initProperty(&queryFields, &mutationFields)
	initOffer(&queryFields, &mutationFields)
	initDocument(&queryFields, &mutationFields)
	initRightmove(&queryFields, &mutationFields)
	initZoopla(&queryFields, &mutationFields)
	initContract(&queryFields, &mutationFields)
	initCardPayment(&queryFields, &mutationFields)
	initSignature(&queryFields, &mutationFields)
	initMaintenance(&queryFields, &mutationFields)

	query := graphql.NewObject(graphql.ObjectConfig{
		Name:        "AdminQuery",
		Description: "Admin query root",
		Fields:      queryFields,
	})

	(*rootQueries)["admin"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(graphql.ObjectConfig{
		Name:        "AdminMutation",
		Description: "Admin mutation root",
		Fields:      mutationFields,
	})

	(*rootMutations)["admin"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}

	initOfferEnrichment()
	initMaintenanceEnrichment()
}

func initOfferEnrichment() {
	util.BindOutput(offerData{}, "offer", util.EnrichmentMap{
		"viewing": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.ViewRequest{}, "viewing", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if req.Tenants == nil {
					req.Tenants = &offer.TenantArrayValue{}
				}
				if !ok || len(req.GetTenants().GetValue()) < 1 {
					return &property.ViewRequest{}, nil
				}
				view, err := connections.PropertyService.GetView(p.Context, &property.GetViewRequest{
					User:     req.GetTenants().GetValue()[0].GetUser(),
					Property: req.GetProperty(),
					Id:       req.GetViewing(),
				})
				if err != nil || view == nil || view.GetRequest() == nil {
					return &property.ViewRequest{}, nil
				}
				return view.GetRequest(), nil
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{Id: req.GetProperty()})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
	})
}

func CensoredAccount(acc *account.Account) *account.Account {
	return &account.Account{
		Id:        acc.GetId(),
		CreatedAt: acc.GetCreatedAt(),
		Details:   acc.GetDetails(),
		Contact:   acc.GetContact(),
	}
}

func ForeignMaintenance(user *auth.AuthResponse, mnt *maintenance.Maintenance) bool {
	return user == nil || (mnt.GetLandlordValue() != user.GetAccount().GetIdValue())
}

func initMaintenanceEnrichment() {
	util.BindOutput(maintenance.Maintenance{}, "maintenance", util.EnrichmentMap{
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				mnt, ok := p.Source.(*maintenance.Maintenance)
				if !ok || mnt.GetUserValue() == "" {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: mnt.GetUserValue()})
				if err != nil {
					return &account.Account{}, err
				}
				return CensoredAccount(acc.GetAccount()), nil
			},
		},
		"landlord": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				mnt, ok := p.Source.(*maintenance.Maintenance)
				if !ok || mnt.GetLandlordValue() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: mnt.GetLandlordValue()})
				if err != nil {
					return nil, err
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignMaintenance(user, mnt) {
					return CensoredAccount(acc.GetAccount()), nil
				}
				return acc.GetAccount(), err
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetPropertyValue() == "" {
					return &property.Property{}, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{Id: req.GetPropertyValue()})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
		"tenants": &util.Enrichment{
			Output: func(bind util.BindType) graphql.Output {
				return graphql.NewList(graphql.NewNonNull(util.BindOutput(&property.Tenant{}, "", nil)))
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				prop, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return []*property.Tenant{}, nil
				}

				return prop.GetTenantsValue(), nil
			},
		},
	})
}
