package admin

import (
	"os"

	"github.com/graphql-go/graphql"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	document "gitlab.okta-solutions.com/mashroom/backend/document/api"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
)

func initDocument(query, mutation *graphql.Fields) {
	(*query)["Document"] = &graphql.Field{
		Description: "Get document by id",
		Args:        util.BindArguments(document.GetRequest{}),
		Resolve:     adminQueryGetDocumentResolver,
		Type:        util.BindOutput(document.GetResponse{}, "Document with given id", nil),
	}
	(*query)["allDocuments"] = &graphql.Field{
		Description: "Get documents by filters",
		Args:        util.BindArguments(document.ListRequest{}),
		Resolve:     adminQueryAllDocumentsResolver,
		Type:        util.BindOutput(document.ListResponse{}, "Document list", nil),
	}

	(*mutation)["createDocument"] = &graphql.Field{
		Description: "Create document",
		Args:        util.BindArguments(document.ServiceCreateRequest{}),
		Resolve:     adminMutationCreateDocumentResolver,
		Type:        util.BindOutput(document.CreateResponse{}, "Document create response", nil),
	}
	(*mutation)["updateDocument"] = &graphql.Field{
		Description: "Update document",
		Args:        util.BindArguments(document.ServiceUpdateRequest{}),
		Resolve:     adminMutationUpdateDocumentResolver,
		Type:        util.BindOutput(document.ServiceUpdateResponse{}, "Document update response", nil),
	}
	(*mutation)["updateDocumentState"] = &graphql.Field{
		Description: "Update document state",
		Args:        util.BindArguments(document.UpdateStatusRequest{}),
		Resolve:     adminMutationUpdateDocumentStateResolver,
		Type:        util.BindOutput(document.UpdateStatusResponse{}, "Document state update response", nil),
	}
	(*mutation)["fileSoftDelete"] = &graphql.Field{
		Description: "File soft delete",
		Args:        util.BindArguments(document.FileSoftDeleteRequest{}),
		Resolve:     adminMutationFileSoftDeleteResolver,
		Type:        util.BindOutput(document.FileSoftDeleteResponse{}, "File soft delete response", nil),
	}
	(*mutation)["cleanThumbsCache"] = &graphql.Field{
		Description: "Clean thumbs cache",
		Args:        util.BindArguments(document.CleanThumbsCacheRequest{}),
		Resolve:     adminMutationCleanThumbsCacheResolver,
		Type:        util.BindOutput(document.CleanThumbsCacheResponse{}, "Clean thumbs cache response", nil),
	}
}

func adminQueryGetDocumentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	doc, err := connections.DocumentService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}
	return doc.GetDocument(), err
}

func adminQueryAllDocumentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if request.GetProperty() != "" {
		resp, err := connections.PropertyService.Get(params.Context, &property.GetRequest{
			Id: request.GetProperty(),
		})
		if err == nil {
			request.User = resp.GetProperty().GetPeople().GetLandlord().GetValue()
		}
	}
	docs, err := connections.DocumentService.List(params.Context, request)
	if err != nil {
		return nil, err
	}

	return docs.GetDocuments(), nil
}

func adminMutationCreateDocumentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.ServiceCreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.User == "" {
		reqProp := &property.GetRequest{
			Id: request.Property,
		}
		prop, err := connections.PropertyService.GetAdminVerify(params.Context, reqProp)
		if err != nil {
			return nil, err
		}
		people := prop.GetProperty().GetPeople()
		if people.GetLandlord() != nil {
			request.User = people.GetLandlord().GetValue()
		} else if people.GetManager() != nil {
			request.User = people.GetManager().GetValue()
		}
	}

	return connections.DocumentService.ServiceCreate(params.Context, request)
}

func adminMutationUpdateDocumentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.ServiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.User == "" {
		reqProp := &property.GetRequest{
			Id: request.Property,
		}
		prop, err := connections.PropertyService.GetAdminVerify(params.Context, reqProp)
		if err != nil {
			return nil, err
		}
		people := prop.GetProperty().GetPeople()
		if people.GetLandlord() != nil {
			request.User = people.GetLandlord().GetValue()
		} else if people.GetManager() != nil {
			request.User = people.GetManager().GetValue()
		}
	}

	return connections.DocumentService.ServiceUpdate(params.Context, request)
}

func adminMutationUpdateDocumentStateResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.UpdateStatusRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.UpdateStatus(params.Context, request)
}

func adminMutationFileSoftDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.FileSoftDeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.DocumentService.FileSoftDelete(params.Context, request)
}

func adminMutationCleanThumbsCacheResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &document.CleanThumbsCacheRequest{}

	resp, err := connections.DocumentService.CleanThumbsCache(params.Context, request)
	if err != nil {
		return nil, err
	}
	err = os.RemoveAll(MediaCacheDir + "/thumb")
	if err != nil {
		return nil, err
	}

	return resp, nil
}
