package admin

import (
	"context"
	"errors"

	"github.com/graphql-go/graphql"

	account "gitlab.okta-solutions.com/mashroom/backend/account/api"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	q "gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/wrap"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers"
	offer "gitlab.okta-solutions.com/mashroom/backend/offer/api"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
)

type allOffersResult struct {
	Data  []*offer.Offer `json:"data,omitempty"`
	Total int32          `json:"total,omitempty"`
}

type OfferFull struct {
	offer.Offer
	Viewing  *property.ViewRequest `json:"viewing,omitempty"`
	Property *property.Property    `json:"property,omitempty"`
}

type offerData struct {
	OfferFull
	Offers   []*offer.Offer          `json:"offers,omitempty"`
	Viewings []*property.ViewRequest `json:"viewings,omitempty"`
}

type offerUpdateResponse struct {
	Status offer.OfferStatus  `json:"status,omitempty"`
	Errors []offer.OfferError `json:"errors,omitempty"`
	Offer  *offerData         `json:"offer,omitempty"`
}

var (
	errPropertyNotFound = errors.New("property not found")
	errLandlordNotFound = errors.New("landlord not found")
)

func initOffer(query, mutation *graphql.Fields) {
	(*query)["Offer"] = &graphql.Field{
		Description: "Retrieves specified offer",
		Args:        util.BindArguments(offer.GetRequest{}),
		Resolve:     adminQueryGetOfferResolver,
		Type:        util.BindOutput(offerData{}, "Offer details", nil),
	}
	(*query)["allOffers"] = &graphql.Field{
		Description: "Retrieves offer list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllOffersResolver,
		Type:        util.BindOutput(allOffersResult{}, "Offer list", nil),
	}

	(*mutation)["updateOffer"] = &graphql.Field{
		Description: "Updates existing offer",
		Args:        util.BindArguments(offer.ServiceUpdateRequest{}),
		Resolve:     adminMutationUpdateOfferResolver,
		Type:        util.BindOutput(offerUpdateResponse{}, "Update status", nil),
	}

	(*mutation)["acceptOffer"] = &graphql.Field{
		Description: "Accepts offer on behalf of a Landlord",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationAcceptOfferResolver,
		Type:        util.BindOutput(offer.AcceptResponse{}, "Accept status", nil),
	}

	(*mutation)["approveOffer"] = &graphql.Field{
		Description: "Approves prepared offer",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationApproveOfferResolver,
		Type:        util.BindOutput(offer.ApproveResponse{}, "Approve status", nil),
	}

	(*mutation)["OfferPaymentStatus"] = &graphql.Field{
		Description: "Updates (confirms) offer payment by tenant via bank",
		Args:        util.BindArguments(offer.PaymentStatusRequest{}),
		Resolve:     adminMutationOfferPaymentStatusResolver,
		Type:        util.BindOutput(offer.PaymentStatusResponse{}, "Offer payment status", nil),
	}

	(*mutation)["RenderOfferCsv"] = &graphql.Field{
		Description: "Generate csv with Tenant and Landlord details to send to the insurer",
		Args:        util.BindArguments(offer.RenderCsvRequest{}),
		Resolve:     adminMutationRenderCsvResolver,
		Type:        util.BindOutput(offer.RenderCsvResponse{}, "Render csv response", nil),
	}

	(*mutation)["executeOfferCheck"] = &graphql.Field{
		Description: "Execute tenants check for offer",
		Args:        util.BindArguments(offer.ExecuteCheckRequest{}),
		Resolve:     adminMutationCheckOfferResolver,
		Type:        util.BindOutput(offer.CheckResponse{}, "Check offer tenants response", nil),
	}

}

func findLandlord(ctx context.Context, landlord *offer.Landlord) *account.Account {
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
	}
	if (landlord.GetEmailPrimary() != nil) && (landlord.GetEmailPrimary().GetValue() != "") {
		req.Filter = append(req.Filter, &q.QueryParamsFilter{
			Field:    "contact.email_primary",
			Value:    []string{landlord.GetEmailPrimary().GetValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}
	if (landlord.GetPhonePrimary() != nil) && (landlord.GetPhonePrimary().GetValue() != "") {
		req.Filter = append(req.Filter, &q.QueryParamsFilter{
			Field:    "contact.phone_primary",
			Value:    []string{landlord.GetPhonePrimary().GetValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}
	request := &account.AccountListFilterRequest{
		Params: &account.AccountListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.Admin = false
	resp, err := connections.AccountService.ListFilter(ctx, request)
	if (err != nil) || (len(resp.GetAccounts()) != 1) {
		return nil
	}
	return resp.GetAccounts()[0]
}

func getViewing(ctx context.Context, off *offer.Offer) (*property.ViewRequest, error) {
	if off.GetTenants() == nil {
		off.Tenants = &offer.TenantArrayValue{}
	}
	if len(off.GetTenants().GetValue()) < 1 {
		return &property.ViewRequest{}, nil
	}
	view, err := connections.PropertyService.GetView(ctx, &property.GetViewRequest{
		User:     off.GetTenants().GetValue()[0].GetUser(),
		Property: off.GetProperty(),
		Id:       off.GetViewing(),
	})
	if err != nil || view == nil || view.GetRequest() == nil {
		return &property.ViewRequest{}, nil
	}
	return view.GetRequest(), nil
}

func getProperty(ctx context.Context, off *offer.Offer) (*property.Property, error) {
	if off.GetProperty() == "" {
		return &property.Property{}, nil
	}
	resp, err := connections.PropertyService.Get(ctx, &property.GetRequest{Id: off.GetProperty()})
	if err != nil || resp == nil || resp.GetProperty() == nil {
		return &property.Property{}, nil
	}
	return resp.GetProperty(), nil
}

func enrichOfferData(ctx context.Context, off *offer.Offer, resp *offerData) {
	if off == nil {
		return
	}
	// Get offers for property from this offer (maybe filter by tenant too?)
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{
			{
				Field:    "property",
				Value:    []string{off.GetProperty()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
		},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
	}
	request := &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.AdminSite = true
	if offs, err := connections.OfferService.ListFilter(ctx, request); err == nil {
		resp.Offers = offs.Offers
	}

	if off.GetLandlord() == nil {
		off.Landlord = &offer.Landlord{}
	}
	// Get view requests list for property from this offer
	landlord := findLandlord(ctx, off.GetLandlord())
	if landlord != nil {
		request := &property.ListViewRequest{
			Property: wrap.StringValue(&off.Property),
			User:     landlord.GetIdValue(),
		}
		viewings, err := connections.PropertyService.ListView(ctx, request)
		if (err == nil) && (viewings.Success) {
			resp.Viewings = viewings.List
		}
	}

	view, err := getViewing(ctx, off)
	if (err == nil) && (view != nil) {
		resp.Viewing = view
	}

	prop, err := getProperty(ctx, off)
	if (err == nil) && (prop != nil) {
		resp.Property = prop
	}

}

func adminQueryGetOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	offr, err := connections.OfferService.ServiceGet(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *offerData = nil
	if offr.GetOffer() != nil {
		resp = &offerData{
			OfferFull: OfferFull{
				Offer: *offr.GetOffer(),
			},
		}
	}
	if (offr != nil) && (resp != nil) {
		enrichOfferData(params.Context, offr.GetOffer(), resp)
	}
	return resp, nil
}

func adminQueryAllOffersResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}
	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created_at", Order: -1}}
	}
	request := &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.AdminSite = true
	resp, err := connections.OfferService.ListFilter(params.Context, request)
	if err != nil {
		return nil, err
	}
	offers := resp.GetOffers()
	total := resp.GetTotal()
	return allOffersResult{
		Data:  offers,
		Total: total,
	}, nil
}

func adminMutationUpdateOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.ServiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.OfferService.ServiceUpdate(params.Context, request)
	if err != nil {
		return nil, err
	}
	var offr *offerData = nil
	if resp.GetOffer() != nil {
		offr = &offerData{
			OfferFull: OfferFull{
				Offer: *resp.GetOffer(),
			},
		}
	}
	if (resp != nil) && (offr != nil) {
		enrichOfferData(params.Context, resp.GetOffer(), offr)
	}
	return &offerUpdateResponse{
		Status: resp.Status,
		Errors: resp.Errors,
		Offer:  offr,
	}, nil
}

func adminMutationAcceptOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	// TODO: Rework this piece...
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	reqOffer := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, reqOffer); err != nil {
		return nil, err
	}
	reqOffer.User = user.GetAccount().GetIdValue()
	offr, err := connections.OfferService.ServiceGet(params.Context, reqOffer)
	if err != nil {
		return nil, err
	}
	if (offr.GetOffer() == nil) || (offr.GetOffer().GetProperty() == "") {
		return nil, errPropertyNotFound
	}

	reqProp := &property.GetRequest{
		Id:   offr.GetOffer().GetProperty(),
		User: reqOffer.User,
	}
	reqProp.User = user.GetAccount().GetIdValue()
	if err = util.UnmarshalArguments(params.Args, reqProp); err != nil {
		return nil, err
	}
	prop, err := connections.PropertyService.Get(params.Context, reqProp)
	if err != nil {
		return nil, err
	}
	if (prop.GetProperty() == nil) ||
		(prop.GetProperty().GetPeople() == nil) ||
		(prop.GetProperty().GetPeople().GetLandlord() == nil) ||
		(prop.GetProperty().GetPeople().GetLandlord().GetValue() == "") {
		return nil, errLandlordNotFound
	}
	landlord := prop.GetProperty().GetPeople().GetLandlord().GetValue()

	request := &offer.AcceptRequest{
		User: landlord,
		Id:   reqOffer.Id,
	}
	return connections.OfferService.Accept(params.Context, request)
}

func adminMutationApproveOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &offer.ApproveRequest{
		User:  user.GetAccount().GetIdValue(),
		Offer: req.ID,
	}
	offr, err := connections.OfferService.ServiceApprove(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *offerData = nil
	if offr.GetOffer() != nil {
		resp = &offerData{
			OfferFull: OfferFull{
				Offer: *offr.GetOffer(),
			},
		}
	}
	if (offr != nil) && (resp != nil) {
		enrichOfferData(params.Context, offr.GetOffer(), resp)
	}
	return resp, nil
}

func adminMutationOfferPaymentStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.PaymentStatusRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.OfferService.ServicePaymentStatus(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func adminMutationRenderCsvResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.RenderCsvRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.OfferService.RenderCsv(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func adminMutationCheckOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.ExecuteCheckRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.OfferService.ExecuteCheck(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}
