package admin

import (
	"context"

	"github.com/graphql-go/graphql"

	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	cardpayment "gitlab.okta-solutions.com/mashroom/backend/cardpayment/api"
	q "gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	document "gitlab.okta-solutions.com/mashroom/backend/document/api"
	extras "gitlab.okta-solutions.com/mashroom/backend/extras/api"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
)

type allPropertiesResult struct {
	Data  []*property.Property `json:"data,omitempty"`
	Total int32                `json:"total,omitempty"`
}

type propertyData struct {
	*property.Property
	Documents   []*document.Document   `json:"documents,omitempty"`
	ListingFee  []*cardpayment.Receipt `json:"listing_fee,omitempty"`
	AdminVerify *property.AdminVerify  `json:"admin_verify,omitempty"`
	Viewings    []*property.Viewing    `json:"viewings,omitempty"`
}

type publishResponse struct {
	Success  bool                     `json:"success,omitempty"`
	Errors   []property.PropertyError `json:"errors,omitempty"`
	ErrorMsg string                   `json:"error_msg,omitempty"`
	Property *propertyData            `json:"property,omitempty"`
}

type unpublishResponse struct {
	Success  bool                     `json:"success,omitempty"`
	Errors   []property.PropertyError `json:"errors,omitempty"`
	ErrorMsg string                   `json:"error_msg,omitempty"`
}

type withdrawResponse struct {
	Success  bool                     `json:"success,omitempty"`
	Errors   []property.PropertyError `json:"errors,omitempty"`
	ErrorMsg string                   `json:"error_msg,omitempty"`
}

func initProperty(query, mutation *graphql.Fields) {
	(*query)["Property"] = &graphql.Field{
		Description: "Retrieves specified property",
		Args:        util.BindArguments(property.GetRequest{}),
		Resolve:     adminQueryGetPropertyResolver,
		Type:        util.BindOutput(propertyData{}, "Property information", nil),
	}
	(*query)["allProperties"] = &graphql.Field{
		Description: "Retrieves property list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllPropertiesResolver,
		Type:        util.BindOutput(allPropertiesResult{}, "Property list", nil),
	}

	(*mutation)["publishProperty"] = &graphql.Field{
		Description: "Publish property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationPublishPropertyResolver,
		Type:        util.BindOutput(publishResponse{}, "Publish property response", nil),
	}
	(*mutation)["unpublishProperty"] = &graphql.Field{
		Description: "Unpublish property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationUnpublishPropertyResolver,
		Type:        util.BindOutput(publishResponse{}, "Unpublish property response", nil),
	}
	(*mutation)["updateProperty"] = &graphql.Field{
		Description: "Update property",
		Args:        util.BindArguments(propertyData{}),
		Resolve:     adminMutationUpdatePropertyResolver,
		Type:        util.BindOutput(propertyData{}, "Property information", nil),
	}
	(*mutation)["reorderPhotos"] = &graphql.Field{
		Description: "Reorder photos",
		Args:        util.BindArguments(property.ReorderPhotosRequest{}),
		Resolve:     adminMutationReorderPhotosResolver,
		Type:        util.BindOutput(property.ReorderPhotosResponse{}, "Reorder information", nil),
	}
	(*mutation)["withdrawProperty"] = &graphql.Field{
		Description: "Withdraw property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationWithdrawPropertyResolver,
		Type:        util.BindOutput(withdrawResponse{}, "Withdraw property response", nil),
	}
	(*mutation)["updateRewardStatus"] = &graphql.Field{
		Description: "Update reward status of the viewing",
		Args:        util.BindArguments(property.NewRewardViewingRequest{}),
		Resolve:     adminMutationUpdateRewardStatusResolver,
		Type:        util.BindOutput(property.NewRewardViewingResponse{}, "Update reward status response", nil),
	}
}

func enrichPropertyData(ctx context.Context, prop *property.Property, resp *propertyData) {
	if prop == nil {
		return
	}

	// Get documents for property
	reqDoc := &document.ListRequest{
		Property: prop.GetId().GetValue(),
		User:     prop.GetPeople().GetLandlord().GetValue(),
	}
	if docs, err := connections.DocumentService.List(ctx, reqDoc); err == nil {
		resp.Documents = docs.GetDocuments()
	}

	// Get listing fees for property
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{
			{
				Field:    "property",
				Value:    []string{prop.GetId().GetValue()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
			{
				Field:    "user_id",
				Value:    []string{prop.GetPeople().GetLandlord().GetValue()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
		},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
	}
	request := &cardpayment.ReceiptListFilterRequest{
		Params: &cardpayment.ReceiptListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.Admin = true
	if rcpts, err := connections.CardPaymentService.ListFilter(ctx, request); err == nil {
		for _, charge := range rcpts.Payments {
			var items []*extras.Service
			for _, item := range charge.Items {
				items = append(items, item)
			}
			if len(items) > 0 {
				charge.Items = items
				resp.ListingFee = append(resp.ListingFee, charge)
			}
		}
	}

	// Get Viewings
	if views, err := connections.PropertyService.ListAllView(ctx, &property.ListViewRequest{
		Property: prop.GetId(),
	}); err == nil {
		resp.Viewings = views.List
	}
}

func adminMutationUpdateRewardStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &property.NewRewardViewingRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.PropertyService.NewRewardViewing(params.Context, request)
}

func adminQueryGetPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &property.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	prop, err := connections.PropertyService.GetAdminVerify(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *propertyData = nil
	if prop.GetProperty() != nil {
		resp = &propertyData{
			Property: prop.GetProperty(),
		}
		if prop.GetAdminVerify() != nil {
			resp.AdminVerify = prop.GetAdminVerify()
		}
	}
	if (prop != nil) && (resp != nil) {
		enrichPropertyData(params.Context, prop.GetProperty(), resp)
	}
	return resp, nil
}

func adminQueryAllPropertiesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}
	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created_at", Order: -1}}
	}
	request := &property.PropertyListFilterRequest{
		Params: req,
	}
	request.AdminSite = true
	resp, err := connections.PropertyService.ListFilter(params.Context, request)
	if err != nil {
		return nil, err
	}
	properties := resp.GetProperties()
	total := resp.GetTotal()
	return allPropertiesResult{
		Data:  properties,
		Total: total,
	}, nil
}

func adminMutationUpdatePropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &propertyData{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.UpdateAndVerifyRequest{
		User:        user.GetAccount().GetIdValue(),
		Property:    req.Property,
		AdminVerify: req.AdminVerify,
	}
	prop, err := connections.PropertyService.UpdateAndVerify(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &propertyData{
		Property:    prop.GetProperty(),
		AdminVerify: prop.GetAdminVerify(),
	}
	if prop != nil {
		enrichPropertyData(params.Context, prop.GetProperty(), resp)
	}
	return resp, nil
}

func adminMutationPublishPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.PublishRequest{
		Id:   req.ID,
		User: user.GetAccount().GetIdValue(),
	}
	r, err := connections.PropertyService.ServicePublish(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &publishResponse{
		Success:  r.GetSuccess(),
		Errors:   r.GetErrors(),
		ErrorMsg: r.ErrorMsg,
		Property: nil,
	}
	prop := r.GetProperty()
	if prop != nil {
		resp.Property = &propertyData{
			Property: prop,
		}
		if r.GetAdminVerify() != nil {
			resp.Property.AdminVerify = r.GetAdminVerify()
		}
	}
	if prop != nil {
		enrichPropertyData(params.Context, prop, resp.Property)
	}
	return resp, nil
}

func adminMutationUnpublishPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.UnpublishRequest{
		Id:        req.ID,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: true,
	}
	r, err := connections.PropertyService.ServiceUnpublish(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &unpublishResponse{
		Success:  r.GetSuccess(),
		Errors:   r.GetErrors(),
		ErrorMsg: r.GetErrorMsg(),
	}
	return resp, nil
}

func adminMutationReorderPhotosResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &property.ReorderPhotosRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.PropertyService.ReorderPhotos(params.Context, req)
}

func adminMutationWithdrawPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.WithdrawnRequest{
		Id:        req.ID,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: true,
	}
	r, err := connections.PropertyService.ServiceWithdrawn(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &withdrawResponse{
		Success:  r.GetSuccess(),
		Errors:   r.GetErrors(),
		ErrorMsg: r.GetErrorMsg(),
	}

	return resp, nil
}
