package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers"
	zoopla "gitlab.okta-solutions.com/mashroom/backend/zoopla/api"
)

type getParamsOpt struct {
	Ids []string `json:"ids"`
}

type deleteZooplaResponse struct {
	Result []*zoopla.DeletePropertyResponse `json:"result"`
}

func initZoopla(query, mutation *graphql.Fields) {
	(*mutation)["publishZoopla"] = &graphql.Field{
		Description: "Publish property on Zoopla",
		Args:        util.BindArguments(getParams{}),
		Resolve:     ZooplaPublishResolver,
		Type:        util.BindOutput(zoopla.SendPropertyResponse{}, "Property ids", nil),
	}
	(*mutation)["deleteZoopla"] = &graphql.Field{
		Description: "Delete published property from Zoopla",
		Args:        util.BindArguments(getParams{}),
		Resolve:     ZooplaDeleteResolver,
		Type:        util.BindOutput(zoopla.DeletePropertyResponse{}, "Delete result", nil),
	}
	(*mutation)["deleteZooplaList"] = &graphql.Field{
		Description: "Delete multiple published property from Zoopla",
		Args:        util.BindArguments(getParamsOpt{}),
		Resolve:     ZooplaDeleteListResolver,
		Type:        util.BindOutput(deleteZooplaResponse{}, "Delete result", nil),
	}
}

func ZooplaPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &zoopla.PropertyRequest{
		PropertyId: req.ID,
		User:       user.GetAccount().GetIdValue(),
	}
	resp, err := connections.ZooplaService.SendProperty(params.Context, request)
	if err != nil {
		return nil, err
	}
	if (resp.GetZooplaId() == "") || (resp.GetZooplaId() == "0") || (resp.GetZooplaUrl() == "") {
		resp.Success = false
	}
	return resp, nil
}

func ZooplaDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &zoopla.DeletePropertyRequest{
		PropertyId:     req.ID,
		User:           user.GetAccount().GetIdValue(),
		IgnoreNotFound: false,
	}
	return connections.ZooplaService.DeleteProperty(params.Context, request)
}

func ZooplaDeleteListResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParamsOpt{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	acc := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	user := acc.GetAccount().GetIdValue()

	var res []*zoopla.DeletePropertyResponse
	for _, id := range req.Ids {
		request := &zoopla.DeletePropertyRequest{
			PropertyId:     id,
			User:           user,
			IgnoreNotFound: true,
		}
		if resp, err := connections.ZooplaService.DeleteProperty(params.Context, request); (err == nil) && (resp != nil) {
			res = append(res, resp)
		}

	}
	return deleteZooplaResponse{Result: res}, nil
}
