package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	cardpayment "gitlab.okta-solutions.com/mashroom/backend/cardpayment/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers/errs"
)

type MarkReceiptPaidRequest struct {
	ID string `json:"id,omitempty"`
}

func initCardPayment(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(graphql.ObjectConfig{
		Name:        "CardPaymentQuery",
		Description: "Card payment query root",
		Fields: graphql.Fields{
			"user": &graphql.Field{
				Description: "Get current user receipts",
				Resolve:     CardPaymentQueryUserResolver,
				Type:        util.BindOutput(cardpayment.GetUserResponse{}, "Receipt list", nil),
			},
			"property": &graphql.Field{
				Description: "Get services list",
				Args:        util.BindArguments(cardpayment.GetPropertyRequest{}),
				Resolve:     CardPaymentQueryPropertyResolver,
				Type:        util.BindOutput(cardpayment.GetPropertyResponse{}, "Receipt list", nil),
			},
		},
	})

	(*rootQueries)["cardpayment"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(graphql.ObjectConfig{
		Name:        "CardPaymentMutation",
		Description: "Card payment mutation root",
		Fields: graphql.Fields{
			"form": &graphql.Field{
				Description: "Form a receipt",
				Args:        util.BindArguments(cardpayment.FormReceiptRequest{}),
				Resolve:     CardPaymentMutationFormResolver,
				Type:        util.BindOutput(cardpayment.FormReceiptResponse{}, "Payment intent id and secret", nil),
			},
			"cancel": &graphql.Field{
				Description: "Cancel a receipt",
				Args:        util.BindArguments(cardpayment.CancelReceiptRequest{}),
				Resolve:     CardPaymentMutationCancelResolver,
				Type:        util.BindOutput(cardpayment.CancelReceiptResponse{}, "Cancellation status", nil),
			},
			"check": &graphql.Field{
				Description: "Check receipt status",
				Args:        util.BindArguments(cardpayment.CheckReceiptRequest{}),
				Resolve:     CardPaymentMutationCheckResolver,
				Type:        util.BindOutput(cardpayment.CheckReceiptResponse{}, "Actual state of receipt", nil),
			},
			"mark": &graphql.Field{
				Description: "Mark as paid",
				Args:        util.BindArguments(cardpayment.MarkReceiptPaidRequest{}),
				Resolve:     CardPaymentMutationMarkResolver,
				Type:        util.BindOutput(cardpayment.MarkReceiptPaidResponse{}, "", nil),
			},
			"markbank": &graphql.Field{
				Description: "Mark as paid via bank",
				Args:        util.BindArguments(cardpayment.MarkReceiptPaidRequest{}),
				Resolve:     CardPaymentMutationMarkBankResolver,
				Type:        util.BindOutput(cardpayment.MarkReceiptPaidResponse{}, "", nil),
			},
		},
	})

	(*rootMutations)["cardpayment"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func CardPaymentQueryUserResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.GetUserRequest{}
	request.User = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.GetUser(params.Context, request)
}

func CardPaymentQueryPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.GetPropertyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.GetProperty(params.Context, request)
}

func CardPaymentMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.CancelReceiptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.CancelReceipt(params.Context, request)
}

func CardPaymentMutationCheckResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.CheckReceiptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.CheckReceipt(params.Context, request)
}

func CardPaymentMutationFormResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.FormReceiptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()
	request.UserEmail = user.GetAccount().GetContactValue().GetEmailPrimaryValue()
	return connections.CardPaymentService.FormReceipt(params.Context, request)
}

func CardPaymentMutationMarkResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &cardpayment.MarkReceiptPaidRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.UserId = user.GetAccount().GetIdValue()
	return connections.CardPaymentService.MarkReceiptPaid(params.Context, request)
}

func CardPaymentMutationMarkBankResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	req := &MarkReceiptPaidRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	request := &cardpayment.MarkReceiptPaidRequest{
		Id:     req.ID,
		UserId: user.GetAccount().GetIdValue(),
	}
	return connections.CardPaymentService.MarkReceiptPaid(params.Context, request)
}
