package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers/errs"
	maintenance "gitlab.okta-solutions.com/mashroom/backend/maintenance/api"
)

func initMaintenance(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(graphql.ObjectConfig{
		Name:        "MaintenanceQuery",
		Description: "Maintenance query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Retrieves specified maintenance",
				Args:        util.BindArguments(maintenance.GetRequest{}),
				Resolve:     MaintenanceQueryGetResolver,
				Type:        util.BindOutput(maintenance.GetResponse{}, "Maintenance information", nil),
			},
			"list": &graphql.Field{
				Description: "Retrieves owned maintenance",
				Args:        util.BindArguments(maintenance.ListRequest{}),
				Resolve:     MaintenanceQueryListResolver,
				Type:        util.BindOutput(maintenance.ListResponse{}, "Maintenance list information", nil),
			},
			"incoming": &graphql.Field{
				Description: "Retrieves owned maintenance",
				Args:        util.BindArguments(maintenance.ListRequest{}),
				Resolve:     MaintenanceQueryListIncomingResolver,
				Type:        util.BindOutput(maintenance.ListResponse{}, "Maintenance list information", nil),
			},
			"outgoing": &graphql.Field{
				Description: "Retrieves owned maintenance",
				Args:        util.BindArguments(maintenance.ListRequest{}),
				Resolve:     MaintenanceQueryListOutgoingResolver,
				Type:        util.BindOutput(maintenance.ListResponse{}, "Maintenance list information", nil),
			},
			"getsubcategories": &graphql.Field{
				Description: "Get subcategories of local heroes",
				Args:        util.BindArguments(maintenance.GetSubcategoryListRequest{}),
				Resolve:     MaintenanceQueryGetSubcategoriesResolver,
				Type:        util.BindOutput(maintenance.GetSubcategoryListResponse{}, "Subcategories list information", nil),
			},
			"getjobdescriptions": &graphql.Field{
				Description: "Get job descriptions of local heroes for subcategory",
				Args:        util.BindArguments(maintenance.GetJobDescriptionListRequest{}),
				Resolve:     MaintenanceQueryGetJobDescriptionsResolver,
				Type:        util.BindOutput(maintenance.GetJobDescriptionListResponse{}, "Job descriptions list information", nil),
			},
		},
	})

	(*rootQueries)["maintenance"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(graphql.ObjectConfig{
		Name:        "MaintenanceMutation",
		Description: "Maintenance mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Creates new maintenance",
				Args:        util.BindArguments(maintenance.CreateRequest{}),
				Resolve:     MaintenanceMutationCreateResolver,
				Type:        util.BindOutput(maintenance.CreateResponse{}, "Create status", nil),
			},
			"update": &graphql.Field{
				Description: "Updates existing maintenance",
				Args:        util.BindArguments(maintenance.UpdateRequest{}),
				Resolve:     MaintenanceMutationUpdateResolver,
				Type:        util.BindOutput(maintenance.UpdateResponse{}, "Update status", nil),
			},
			"assignlh": &graphql.Field{
				Description: "Assign issue to local heroes",
				Args:        util.BindArguments(maintenance.AssignIssueLocalHeroesRequest{}),
				Resolve:     MaintenanceMutationAssignLocalHeroesResolver,
				Type:        util.BindOutput(maintenance.AssignIssueLocalHeroesResponse{}, "Assign status", nil),
			},
		},
	})

	(*rootMutations)["maintenance"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func MaintenanceMutationAssignLocalHeroesResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &maintenance.AssignIssueLocalHeroesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.MaintenanceService.AssignIssueLocalHeroes(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryGetJobDescriptionsResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &maintenance.GetJobDescriptionListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.MaintenanceService.GetJobDescriptionList(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryGetSubcategoriesResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &maintenance.GetSubcategoryListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.MaintenanceService.GetSubcategoryList(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &maintenance.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if user != nil {
		request.User = user.GetAccount().GetIdValue()
	}
	resp, err := connections.MaintenanceService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &maintenance.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.MaintenanceService.List(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryListIncomingResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &maintenance.ListIncomingRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.MaintenanceService.ListIncoming(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceQueryListOutgoingResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &maintenance.ListOutgoingRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.MaintenanceService.ListOutgoing(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &maintenance.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.MaintenanceService.Create(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func MaintenanceMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &maintenance.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.MaintenanceService.Update(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}
