package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.okta-solutions.com/mashroom/backend/auth/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/graphql/util"
	ws "gitlab.okta-solutions.com/mashroom/backend/common/graphql/websocket"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/resolvers/errs"
	offer "gitlab.okta-solutions.com/mashroom/backend/offer/api"
)

func initOffer(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(graphql.ObjectConfig{
		Name:        "OfferQuery",
		Description: "Offer query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Retrieves specified offer",
				Args:        util.BindArguments(offer.GetRequest{}),
				Resolve:     OfferQueryGetResolver,
				Type:        util.BindOutput(offer.GetResponse{}, "Requested offer", nil),
			},
			"listbyproperty": &graphql.Field{
				Description: "Retrieves all offers by property",
				Args:        util.BindArguments(offer.GetRequest{}),
				Resolve:     OfferQueryListByPropertyResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"myout": &graphql.Field{
				Description: "Retrieves current user outgoing offers",
				Args:        util.BindArguments(offer.ListRequest{}),
				Resolve:     OfferQueryMyoutResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"myin": &graphql.Field{
				Description: "Retrieves current user incoming offers",
				Args:        util.BindArguments(offer.ListRequest{}),
				Resolve:     OfferQueryMyinResolver,
				Type:        util.BindOutput(offer.ListResponse{}, "Offer list", nil),
			},
			"serviceslist": &graphql.Field{
				Description: "List services",
				Args:        util.BindArguments(offer.ListServicesRequest{}),
				Resolve:     OfferQueryServicesListResolver,
				Type:        util.BindOutput(offer.ListServicesResponse{}, "Services list", nil),
			},
		},
	})

	(*rootQueries)["offer"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	inviteRequest := util.BindArguments(offer.InviteRequest{})
	delete(inviteRequest, "site_url")

	mutation := graphql.NewObject(graphql.ObjectConfig{
		Name:        "OfferMutation",
		Description: "Offer mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Create new offer",
				Args:        util.BindArguments(offer.CreateRequest{}),
				Resolve:     OfferMutationCreateResolver,
				Type:        util.BindOutput(offer.CreateResponse{}, "Creation status", nil),
			},
			"update": &graphql.Field{
				Description: "Updates existing offer",
				Args:        util.BindArguments(offer.UpdateRequest{}),
				Resolve:     OfferMutationUpdateResolver,
				Type:        util.BindOutput(offer.UpdateResponse{}, "Update status", nil),
			},
			"submit": &graphql.Field{
				Description: "Submit offer for review",
				Args:        util.BindArguments(offer.SubmitRequest{}),
				Resolve:     OfferMutationSubmitResolver,
				Type:        util.BindOutput(offer.SubmitResponse{}, "Submit status", nil),
			},
			"invite": &graphql.Field{
				Description: "Invite tenant to join offer",
				Args:        inviteRequest,
				Resolve:     OfferMutationInviteResolver,
				Type:        util.BindOutput(offer.InviteResponse{}, "Invite creation status", nil),
			},
			"updatetenant": &graphql.Field{
				Description: "Update tenant details",
				Args:        util.BindArguments(offer.UpdateTenantRequest{}),
				Resolve:     OfferMutationUpdatetenantResolver,
				Type:        util.BindOutput(offer.UpdateTenantResponse{}, "Update tenant status", nil),
			},
			"removetenant": &graphql.Field{
				Description: "Remove tenant details",
				Args:        util.BindArguments(offer.RemoveTenantRequest{}),
				Resolve:     OfferMutationRemovetenantResolver,
				Type:        util.BindOutput(offer.RemoveTenantResponse{}, "Remove tenant status", nil),
			},
			"accept": &graphql.Field{
				Description: "Accepts incoming offer",
				Args:        util.BindArguments(offer.AcceptRequest{}),
				Resolve:     OfferMutationAcceptResolver,
				Type:        util.BindOutput(offer.AcceptResponse{}, "Accept status", nil),
			},
			"deny": &graphql.Field{
				Description: "Denies incoming offer",
				Args:        util.BindArguments(offer.DenyRequest{}),
				Resolve:     OfferMutationDenyResolver,
				Type:        util.BindOutput(offer.DenyResponse{}, "Deny status", nil),
			},
			"requestchange": &graphql.Field{
				Description: "Request change for incoming offer",
				Args:        util.BindArguments(offer.RequestChangeRequest{}),
				Resolve:     OfferMutationRequestchangeResolver,
				Type:        util.BindOutput(offer.RequestChangeResponse{}, "Request change status", nil),
			},
			"cancel": &graphql.Field{
				Description: "Cancels own outgoing offer",
				Args:        util.BindArguments(offer.CancelRequest{}),
				Resolve:     OfferMutationCancelResolver,
				Type:        util.BindOutput(offer.CancelResponse{}, "Cancel status", nil),
			},
			"checks": &graphql.Field{
				Description: "Start tenant checks",
				Args:        util.BindArguments(offer.CheckRequest{}),
				Resolve:     OfferMutationCheckResolver,
				Type:        util.BindOutput(offer.CheckResponse{}, "Checks status", nil),
			},
			"payment": &graphql.Field{
				Description: "Request payment url",
				Args:        util.BindArguments(offer.PaymentRequest{}),
				Resolve:     OfferMutationPaymentResolver,
				Type:        util.BindOutput(offer.PaymentResponse{}, "Payment response", nil),
			},
			"servicesselect": &graphql.Field{
				Description: "Select services",
				Args:        util.BindArguments(offer.SelectServicesRequest{}),
				Resolve:     OfferMutationServicesSelectResolver,
				Type:        util.BindOutput(offer.SelectServicesResponse{}, "Services select response", nil),
			},
			"delete": &graphql.Field{
				Description: "Delete offer",
				Args:        util.BindArguments(offer.DeleteOfferByIdRequest{}),
				Resolve:     OfferMutationDeleteResolver,
				Type:        util.BindOutput(offer.DeleteOfferByIdResponse{}, "Delete response", nil),
			},
			"restore": &graphql.Field{
				Description: "Restore offer",
				Args:        util.BindArguments(offer.RestoreOfferByIdRequest{}),
				Resolve:     OfferMutationRestoreResolver,
				Type:        util.BindOutput(offer.RestoreOfferByIdResponse{}, "Restore response", nil),
			},
			"requiredguarantor": &graphql.Field{
				Description: "Request to add a new guarantor to the current tenant",
				Args:        util.BindArguments(offer.RequestRequiredGuarantorRequest{}),
				Resolve:     OfferMutationRequiredGuarantorResolver,
				Type:        util.BindOutput(offer.RequestRequiredGuarantorResponse{}, "Required guarantor response", nil),
			},
			"addguarantors": &graphql.Field{
				Description: "Add new guarantors by tenant",
				Args:        util.BindArguments(offer.AddGuarantorsRequest{}),
				Resolve:     OfferMutationAddGuarantorsResolver,
				Type:        util.BindOutput(offer.AddGuarantorsResponse{}, "Add guarantors response", nil),
			},
		},
	})

	(*rootMutations)["offer"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func OfferMutationRequiredGuarantorResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RequestRequiredGuarantorRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RequestRequiredGuarantor(params.Context, request)
}

func OfferMutationAddGuarantorsResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AddGuarantorsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.AddGuarantors(params.Context, request)
}

func OfferQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Get(params.Context, request)
}

func OfferQueryListByPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.OfferService.ListByProperty(params.Context, request)
}

func OfferQueryMyoutResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ListOutgoing(params.Context, request)
}

func OfferQueryServicesListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListServicesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ServicesList(params.Context, request)
}

func OfferQueryMyinResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ListIncoming(params.Context, request)
}

func OfferMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Create(params.Context, request)
}

func OfferMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Update(params.Context, request)
}

func OfferMutationAcceptResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.AcceptRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Accept(params.Context, request)
}

func OfferMutationDenyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.DenyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Deny(params.Context, request)
}

func OfferMutationCancelResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.CancelRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Cancel(params.Context, request)
}

func OfferMutationSubmitResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.SubmitRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Submit(params.Context, request)
}

func OfferMutationInviteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.InviteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	request.SiteUrl = connections.SiteURL
	return connections.OfferService.Invite(params.Context, request)
}

func OfferMutationUpdatetenantResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.UpdateTenantRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if request.Tenant == nil {
		request.Tenant = &offer.Tenant{}
	}
	request.Tenant.User = user.GetAccount().GetIdValue()
	return connections.OfferService.UpdateTenant(params.Context, request)
}

func OfferMutationRemovetenantResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RemoveTenantRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RemoveTenant(params.Context, request)
}

func OfferMutationRequestchangeResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RequestChangeRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RequestChange(params.Context, request)
}

func OfferMutationCheckResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.CheckRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Check(params.Context, request)
}

func OfferMutationPaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.PaymentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.Payment(params.Context, request)
}

func OfferMutationServicesSelectResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.SelectServicesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.ServicesSelect(params.Context, request)
}

func OfferMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.DeleteOfferByIdRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.DeleteOfferById(params.Context, request)
}

func OfferMutationRestoreResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &offer.RestoreOfferByIdRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.OfferService.RestoreOfferById(params.Context, request)
}
