package rest

import (
	"errors"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"

	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"gitlab.okta-solutions.com/mashroom/backend/gateway/impl/connections"
	offer "gitlab.okta-solutions.com/mashroom/backend/offer/api"
	payment "gitlab.okta-solutions.com/mashroom/backend/payment/api"
	tokenio "gitlab.okta-solutions.com/mashroom/backend/tokenio/api"
)

const errorURL = "https://%s/dashboard"

var errTermsDeclinedRequestIDEmpty = errors.New("terms declined but requestId empty")

func onError(w http.ResponseWriter, err error) {
	log.WithError(err).Error("TokenioHandler")
	w.WriteHeader(http.StatusInternalServerError)
	w.Header().Set("Location", fmt.Sprintf(errorURL, connections.SiteURL))
	_, _ = w.Write(([]byte)(err.Error()))
}

func TokenioHandler(w http.ResponseWriter, r *http.Request) {
	c, err := r.Cookie("csrf_tokenio")
	if err != nil {
		onError(w, err)
		return
	}

	// check decline error
	if len(r.URL.Query()) > 0 {
		for k, v := range r.URL.Query() {
			if strings.ToLower(k) == "error_description" && len(v) > 0 && strings.ToLower(v[0]) == "terms declined" {
				tokenResponse, errT := connections.OfferService.PaymentCsrfToken(r.Context(), &offer.PaymentCsrfTokenRequest{
					CsrfToken: c.Value,
				})
				if errT != nil {
					onError(w, errT)
					return
				}

				offr, errOf := connections.OfferService.ServiceGet(r.Context(), &offer.GetRequest{
					Id: tokenResponse.Offer,
				})
				backLink := fmt.Sprintf("https://%s/dashboard", connections.SiteURL)
				if errOf != nil {
					log.WithError(err).Info("TokenioHandler error get offer details")
				} else {
					if offr.GetOffer().State == offer.OfferState_SIGNED {
						backLink = fmt.Sprintf("https://%s/dashboard/rent?offer=%s", connections.SiteURL, tokenResponse.Offer)
					} else {
						backLink = fmt.Sprintf("https://%s/dashboard/deposit?offer=%s", connections.SiteURL, tokenResponse.Offer)
					}
				}

				w.Header().Set("Location", backLink)
				w.WriteHeader(http.StatusFound)
				return
			}
		}
	}

	cbURL := fmt.Sprintf("https://%s%s", connections.SiteURL, r.URL.String())
	log.WithField("CallbackUrl", cbURL).WithField("CsrfToken", c.Value).Info()
	resp, err := connections.PaymentService.RedeemToken(r.Context(), &payment.RedeemTokenRequest{
		CallbackUrl: cbURL,
		CsrfToken:   c.Value,
	})
	if err != nil {
		onError(w, err)
		return
	}
	log.WithField("ID", resp.GetRefId()).WithField("Status", resp.GetStatus()).WithField("TransferID", resp.GetTransferId()).Info()
	oResp, err := connections.OfferService.PaymentResult(r.Context(), &offer.PaymentResultRequest{
		Id:         resp.GetRefId(),
		Status:     resp.GetStatus(),
		TransferId: resp.GetTransferId(),
	})
	if err != nil {
		onError(w, err)
		return
	}
	log.WithField("Offer", oResp.GetOffer()).WithField("Ok", oResp.GetOk())

	offerResp, errOffr := connections.OfferService.ServiceGet(r.Context(), &offer.GetRequest{
		Id:     oResp.GetOffer(),
		Verify: false,
	})
	if errOffr != nil {
		onError(w, errOffr)
		return
	}
	log.WithField("TokenioHandler success link", offerResp.GetOffer())
	w.Header().Set("Location", fmt.Sprintf("https://%s/dashboard/%s/overview", connections.SiteURL, offerResp.GetOffer().GetProperty()))

	w.WriteHeader(http.StatusFound)
}

func TokenioWebhookHandler(w http.ResponseWriter, r *http.Request) {
	// check decline error
	if len(r.URL.Query()) > 0 {
		errorStr := r.URL.Query().Get("error_description")
		if errorStr != "" && strings.ToLower(errorStr) == "terms declined" {
			token := r.URL.Query().Get("requestId")
			if token != "" {
				tResp, err := connections.OfferService.PaymentCsrfToken(r.Context(), &offer.PaymentCsrfTokenRequest{
					CsrfToken: token,
				})
				if err != nil {
					onError(w, err)
					return
				}

				offr, err := connections.OfferService.ServiceGet(r.Context(), &offer.GetRequest{
					Id: tResp.Offer,
				})
				backLink := fmt.Sprintf("https://%s/dashboard", connections.SiteURL)
				if err != nil {
					log.WithError(err).Info("TokenioHandler error get offer details")
				} else {
					if offr.GetOffer().State == offer.OfferState_SIGNED {
						backLink = fmt.Sprintf("https://%s/dashboard/rent?offer=%s", connections.SiteURL, tResp.Offer)
					} else {
						backLink = fmt.Sprintf("https://%s/dashboard/deposit?offer=%s", connections.SiteURL, tResp.Offer)
					}
				}

				w.Header().Set("Location", backLink)
				w.WriteHeader(http.StatusFound)
			}
			onError(w, errTermsDeclinedRequestIDEmpty)
			return
		}
	}

	body, err := ioutil.ReadAll(r.Body)

	if len(body) == 0 && r.URL != nil {
		body = []byte(r.URL.RawQuery)
	}
	if len(body) == 0 {
		onError(w, err)
		return
	}

	resp, err := connections.TokenioService.PaymentHandler(r.Context(), &tokenio.PaymentHandlerRequest{
		Data: body,
	})
	if err != nil {
		onError(w, err)
		return
	}

	log.WithField("ID", resp.GetRefNum()).WithField("Status", resp.GetStatus()).WithField("TransferID", resp.GetTransferId()).Info()
	oresp, err := connections.OfferService.PaymentResult(r.Context(), &offer.PaymentResultRequest{
		Id:         resp.GetRefNum(),
		Status:     resp.GetStatus(),
		TransferId: resp.GetTransferId(),
	})
	if err != nil {
		onError(w, err)
		return
	}
	log.WithField("Offer", oresp.GetOffer()).WithField("Ok", oresp.GetOk())

	offerResp, errOffr := connections.OfferService.ServiceGet(r.Context(), &offer.GetRequest{
		Id:     oresp.GetOffer(),
		Verify: false,
	})
	if errOffr != nil {
		onError(w, errOffr)
		return
	}

	log.WithField("TokenioHandler success link", offerResp.GetOffer())
	w.Header().Set("Location", fmt.Sprintf("https://%s/dashboard/%s/overview", connections.SiteURL, offerResp.GetOffer().GetProperty()))
	w.WriteHeader(http.StatusFound)
}
